/*******************************************************************************
 *	BdsDataFileBds.h	Bds internal data file format
 *				T.Barnaby,	BEAM Ltd,	2008-05-06
 *******************************************************************************
 *
 *	Supported sub-formats:
 *		BDS, BDS-SM	BDS Sample Multiplexed
 *		BDS-CM		BDS Channel Multiplexed
 *		BDS-SM-CC	BDS Sample Multiplexed, Canadian Compression
 *
 * A BDS data file contains multiple streams. Each stream, numbered from 0 upwards, contains information blocks or
 *  data blocks for a specic data channel or set of channels if the channel data is sample multiplexed into a block.
 * Data streams are split into segments at time discontinuities when using the getInfo() function..
 * Each segment consists of a set of data blocks and each data block can be for one channel or a number of sample multiplexed channels.
 * The siesmic channel to stream mapping is stored in the Information blocks. An information block is located at the start of the
 *  file and at each append point. You can only append data channels that the files was originally created for (ie. the first setInfo() API call).
 *
 * If the ReadOptionReorder option is used the data segments and blocks are sorted into time order.
 *
 * The file can be appended to. When opened for appending
 */
#ifndef BdsDataFileBds_H
#define BdsDataFileBds_H	1

#include <BdsDataFile.h>
#include <BBuffer.h>

namespace Bds {

/// BdsDataFileBds: internal file block type field
enum BdsDataType		{ BdsDataTypeBlock = 0x42534442, BdsDataTypeInfo = 0x30534442, BdsDataTypeData = 0x31534442, BdsDataTypeInfoExtra = 0x32534442 };

/// BdsDataFileBds: internal fixed size BDS Data Block header
struct BdsDataBlockHeader {
	BUInt32		type;			///< Blocks type, contains magic number for synchronisation
	BUInt32		length;			///< Packet length
	BUInt32		packetOffset;		///< Offset to first packet header within block or 0 if no packet header within block
};

/// BdsDataFileBds: internal fixed size BDS Data Block
struct BdsDataBlock {
	BdsDataBlockHeader	header;
	char			data[4];	///< The packet data
};

/// BdsDataFileBds internal file storage packet header
struct BdsDataPacketHeader {
	BUInt32		type;			///< Packets type
	BUInt32		length;			///< Length in bytes of packet
	BUInt32		streamlet;		///< The streamlet id
	BUInt32		sequence;		///< The streamlet packet sequence number
	BUInt32		checksum;		///< Checksum of packet
	BTimeStamp	startTime;		///< The time of the first sample
	BTimeStamp	endTime;		///< The time of the last sample + 1
};

/// BdsDataFileBds: internal file storage packet
class BdsDataPacket : public BBufferStore {
public:
			BdsDataPacket();
			~BdsDataPacket();

	void		clear();
	void		reset();

	void		setChecksumAndLength();
	BError		validateChecksum();

	BError		setHeader(const BdsDataPacketHeader& header);
	BError		getHeader(BdsDataPacketHeader& header);
	
	void		dump();
};

/// BdsDataFileBds: internal file storage data block position
class BdsDataBlockPos {
public:
				BdsDataBlockPos(BTimeStamp startTime = 0, BTimeStamp endTime = 0, BUInt32 channel = 0, BUInt32 numChannels = 0, BUInt32 segment = 0, BUInt64 position = 0, BUInt64 numSamples = 0) :
					startTime(startTime), endTime(endTime), channel(channel), numChannels(numChannels), segment(segment), position(position), numSamples(numSamples){}
	int			operator<(const BdsDataBlockPos& b) const { return startTime < b.startTime; }

	BTimeStamp		startTime;
	BTimeStamp		endTime;
	BUInt32			channel;
	BUInt32			numChannels;
	BUInt32			segment;
	BUInt64			position;
	BUInt64			numSamples;
};

/// BdsDataFileBds: internal file storage data segment
class BdsDataSegment {
public:
				BdsDataSegment() : numBlocks(0), numSamples(0), sampleRate(0.0){};
	int			operator<(const BdsDataSegment& b) const { return startTime < b.startTime; }

	BTimeStamp		startTime;
	BTimeStamp		endTime;
	BUInt32			numBlocks;
	BUInt32			numSamples;
	double			sampleRate;
	BArray<BdsDataBlockPos>	blocks;			// The list of blocks in the segment
};

/// BdsDataFileBds: internal file storage data streamlet
class BdsDataStreamlet {
public:
				BdsDataStreamlet() : packetNumber(0), position(0), channel(0), numChannels(0){}

	BUInt32			packetNumber;
	BUInt64			position;
	BUInt32			channel;		// The base channel number of data in this streamlet
	BUInt32			numChannels;		// The number of channels of data in this streamlet
	BArray<BdsDataBlockPos>	blocks;			// All blocks in this streamlet
	BArray<BdsDataSegment>	segments;		// Segments of data
};


/// This class implements the BDS Data File/Stream access system
class DataFileBds : public DataFile {
public:
	enum			{ StreamsMax = 256 };
	enum			{ DefaultBlockSize = 65536 };
	enum PackFormat		{ PackFormat_Unknown = 0, PackFormat_SM = 1, PackFormat_CM = 2, PackFormat_SM_CC = 3 };
	
	// Generic access
				DataFileBds();
				~DataFileBds();
	
	BError			open(BString fileName, BString mode);							///< Open the file for reading or writing
	BError			flush();					///< Flush any data to disk even if blocks are not full
	BError			close();										///< Close file

				// Write routines
	BError			setFormat(BString format);								///< Sets the sub-format
	BError			setInfo(const DataInfo& dataInfo, const ChannelInfos& channelInfos, WriteOptionsList options = WriteOptionSensorData);		///< Sets the information
	BError			writeData(const DataBlock& data);								///< Writes a data block to the file

				// Read routines
	DataOrder		getDataOrder();
	BError			getInfo(DataInfo& dataInfo, DataFileOptions options, BList<DataError>& errors);			///< Get information on open file
	BError			seekBlock(BUInt32 channel, BUInt segment, BTimeStamp time, BUInt32& blockNumber, BUInt64& sampleNumber, DataBlock& dataBlock);	///< Find the block that contains the samples for the time requested
	BError			readData(BUInt32 channel, BUInt segment, BUInt32 blockNumber, DataBlock& dataBlock);		///< Read the data block for the given channel or all channels if blockNumber is 0

	static DataFormat	getFormats();					///< Get the names of the supported formats

	BError			setDiskBlockSize(BUInt32 blockSize);				///< Sets up file/stream block size
	BUInt32			getDiskBlockSize();						///< Returns the data block size in bytes
	
private:
	void			clear();

	// DataBlock exists checks
	BError			dataBlockAdd(BUInt32 streamlet, BTimeStamp startTime, BTimeStamp endTime);
	Bool			dataBlockCheckIfExists(BUInt32 streamlet, BTimeStamp startTime, BTimeStamp endTime);		///< Checks if this data block already exists
	
	// Low level file access
	BError			setInfoRepeat(BUInt32 repeat);								///< Sets the reapeat period for the information packets
	BError			infoSet(BUInt32 streamlet, BTimeStamp startTime, BTimeStamp endTime, BDictString& info);	///< Sets the information packet and writes this to the file/stream
	BError			infoGet(BdsDataPacket& packet, BDictString& info);					///< Gets the information from an information packet
	BError			readInfoPacket();									///< Read the info packet from the file
	BError			addInfoPacket();
	BError			addInfoExtraPacket();
	BError			streamletFromChannel(BUInt channel, BUInt& streamlet);					///< Find streamlet given channel

	BError			dataAppend(const DataBlock& data);							///< Appends a data block to the file
	BError			dataGet(BUInt channel, BdsDataPacket& packet, DataBlock& data);				///< Gets the data from a data packet

	// Low Level Packet I/O
	BError			setWritePositionForAppend();								///< Sets the next packet write position.
	BError			packetRead(BdsDataPacket& packet);							///< Reads a packet from the file
	BError			packetWrite(BdsDataPacket& packet);							///< Writes a packet to the file
	
	// Low level Disk Block access functions
	BError			diskBlockWrite(void* data, BUInt32 numBytes, int header = 0);	///< Writes data to a block in the file
	BError			diskBlockWriteFlush();						///< Flushes a semi-filled block to the file
	BError			diskBlockRead(void* data, BUInt32 numBytes, int header = 0);	///< Reads data from a block
	BError			diskBlockSeek(BUInt64 position);

	BString			omode;					///< The mode (local)
	BString			oformat;				///< The format	
	PackFormat		opackFormat;				///< The data packet format
	DataInfo		odataInfo;				///< Information on the channel data
	ChannelInfos		ochannelInfos;				///< Channel meta data information
	
	BUInt32			oinfoRepeat;				///< How often to output info packet
	BDictString		oinfo;					///< Info packet output
	BdsDataPacketHeader	oinfoHeader;				///< Info packet header

	// Disk Block access data
	BUInt32			odiskBlockSize;				///< The block size
	BdsDataBlock*		odiskBlockRead;				///< Data block buffer for reads
	BUInt64			odiskPositionRead;			///< The disk block position
	BdsDataBlock*		odiskBlockWrite;			///< Data block buffer for writes
	BUInt64			odiskPositionWrite;			///< The disk block position for writing

	BArray<BdsDataStreamlet>	ostreamlets;			///< Streamlet sequence numbers, position and segments
	BdsDataPacket			opacket;			///< Temporary packet for header/data packet generation
};
}
#endif
