/*******************************************************************************
 *	BdsLib.h	Data File Access
 *			T.Barnaby,	BEAM Ltd,	2008-12-03
 *******************************************************************************
 */
/// @file BdsLib.h
/// @brief General BdsLib API functions
///
#ifndef BdsLib_H
#define	BdsLib_H

#include <BdsD.h>
#include <BdsC.h>
#include <BString.h>

namespace Bds {

const int	NetworkNameLen = 3;		///< Maximum Network name length
const int	StationNameLen = 5;		///< Maximum Station name length
const int	ChannelTypeLen = 3;		///< Maximum Channel type name length
const int	ChannelAuxLen = 2;		///< Maximum Channel Aux length
const int	SourceLen = 16;			///< Maximum Source length

/// Response object adding string conversion
class ResponseObj : public Response {
public:
			ResponseObj(const Response& response);
			~ResponseObj();
	
	BString		getString();
	void		setString(BString str);
};

/// This stores a data error. It includes and error number and a string as well as information on what seismic channel it is for.
class DataError {
public:
			DataError();
			DataError(int errorNumber, BString title, BString filename, BTimeStamp startTime, BTimeStamp endTime, DataInfo& dataInfo, BUInt channel, BString description, BString user = "");
	DataError&	set(int errorNumber, BString title, BString importFilename, BTimeStamp startTime, BTimeStamp endTime, DataInfo& dataInfo, BUInt channel, BString description, BString user = "");
	void		mergeDataInfo(const DataInfo& dataInfo, BUInt channel);

	int		getErrorNumber() const;				///< Get The error number
	BString		getTitle() const;				///< Get the title

	BError		setString(BString str);				///< Set from string
	BError		setStringUser(BString str, BString user);	///< Set from string given by user on command line
	BString		getString() const;				///< Get error message

	int		num() const;					///< Get The error number
	const char*	str() const;					///< Return a char* string
	
			operator int() const;				///< Return error number

public:
	BInt32			oerrorNumber;		///< Error number
	BString			otitle;			///< The title
	BString			odescription;		///< The description
	BString			ofilename;		///< The import filename;
	BTimeStamp		ostartTime;		///< The start Time
	BTimeStamp		oendTime;		///< The end Time
	BString			onetwork;		///< The network Name
	BString			ostation;		///< The station/array name
	BString			ochannel;		///< The channel name
	BString			osource;		///< The data Source
	BString			ouser;			///< The user
};

// Bds Channel name manipulations
void	bdsChannelGetTypeAux(BString name, BString& type, BString& aux);	///< Get the channel type and aux fields from a generic channel name
BString	bdsChannelGetName(BString type, BString aux);				///< Create a full channel name from a channels type and aux fields

// Bds Data Meta data manipulations
BError	bdsDataInfoSetTimeRange(DataInfo& dataInfo);					///< Restricts the time tange of all of the DataInfo's channels to match the DataInfo's startTime/endTime fields
BError	bdsDataInfoFromInfo(BDictString info, DataInfo& dataInfo, Bool append = 0);	///< Sets up a DataInfo object from a BDictString list of named strings
BError	bdsInfoFromDataInfo(const DataInfo& dataInfo, BDictString& info);		///< Converts a DataInfo object into a BDictString list of named strings
BError	bdsDataInfoFlatten(DataInfo& dataInfo);						///< Flattens a DataInfo to 1 segement per channel for use in dataOpen() calls.
BError	bdsDataInfoMergeFlatten(DataInfo& dataInfo, const DataInfo& dataInfoAdd);	///< Merges a DataInfo into another flattening the segments to 1 for use in dataOpen() calls.


BString bdsStationAlias(Station station);							///< Returns the station alias if set else its name
BString	bdsDataChannelInfo(const DataChannel& dataChannel);					///< Returns a string representation of a DataChannel object
BString	bdsDataChannelRef(const DataChannel& dataChannel);					///< Returns the string reference name of a DataChannel object
BString	bdsDataChannelRef(const ChannelInfo& channelInfo);					///< Returns the string reference name of a ChannelInfo object
BError	bdsDataChannelOverallResponse(const ChannelInfo& channelInfo, Response& response);	///< Returns the overal response from the list of responses in a ChannelInfo
BString	bdsSelectionChannelInfo(const Selection& selection, BUInt channel);			///< Returns a string describing the name and time period of a selection channel
BString	bdsFileNameExpand(BString fileName, ChannelInfo& channelInfo);				///< Default filename from a ChannelInfo
BString	bdsFileNameExpand(BString fileName, ChannelInfos& channelInfos);			///< Default filename from a list of ChannelInfo's

// BDS Metadata higher level functionality

// Calculations
double	bdsPoleZeroGain(const PoleZero& poleZero, double frequency);		///< Calculates the overal gain of the given PoleZero transfer function
void	bdsPoleZeroGainPhase(const PoleZero& poleZero, double frequency, double& gain, double& phase);	///< Calculates the overal gain and phase of the given PoleZero transfer function
void	bdsPoleZeroToFap(const PoleZero& poleZero, BUInt nPoints, double calibrationFrequency, double sampleFrequency, BArray<Fap>& fap);	///< Convert PoleZero to FAP

// Bds Debug functions
void	bdsDumpSelection(Selection sel);				///< Debug print out a Selection object
void	bdsDumpDataInfo(DataInfo dataInfo, int includeInfo = 0);	///< Debug print out a DataInfo object
void	bdsDumpChannelInfos(const ChannelInfos& channelInfos);		///< Debug print out a ChannelInfos object
void	bdsDumpData(const DataBlock& dataBlock, int nSamples = 0);	///< Debug print out a DataBlock object
void	bdsDumpPoleZeros(PoleZero poleZeros);				///< Debug print out a PoleZeros object

}
#endif
